#include <qapplication.h>
#include <qwidget.h>
#include <qpaintdevice.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qslider.h>
#include "design/camctrlpanel.h"
#include "moc_camview.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/ioctl.h>

#define PWC_FPS_SHIFT	16
#define PWC_FPS_FRMASK	0x3f0000

#define VIDIOCPWCSAGC _IOW('v', 200, int)
#define VIDIOCPWCSAWB _IOW('v', 202, int)

struct pwc_whitebalance {
  int mode;
  int manual_red, manual_blue;    /* R/W */
  int read_red, read_blue;        /* R/O */
};

CamView::CamView(): initialized(false), panel_shown(false),
	lookup_yuv420p(false), lookup_rgb24(false), rgb_mode(false) {
  ctrl_panel = new CamCtrlPanel( this );
  ctrl_panel->move( 0, height() );
  ctrl_panel->resize( width(), ctrl_panel->height() );
  ctrl_panel->show();
  connect( ctrl_panel->dis_scaleCB, SIGNAL(toggled(bool)), SLOT(disableScale(bool)) );
  connect( ctrl_panel->ratioCB, SIGNAL(toggled(bool)), SLOT(keepRatio(bool)) );
  connect( ctrl_panel->resolCB, SIGNAL(activated(int)), SLOT(setResolution(int)) );
  connect( ctrl_panel->rateCB, SIGNAL(activated(int)), SLOT(setRate(int)) );
  connect( ctrl_panel->autoGainCB, SIGNAL(toggled(bool)), SLOT(setAutoGain(bool)) );
  connect( ctrl_panel->gainSL, SIGNAL(valueChanged(int)), SLOT(setGain(int)) );
 
  if ( !v4lInit() ) return;
  resize( video_xsize, video_ysize );
  show();
  if ( !xvInit() ) return;
  initialized = true;

  resize( 640, 480 );
}

CamView::~CamView() {
  if ( !initialized ) return;
  close( false );
  xvCleanup();
  v4lCleanup();
}

bool CamView::v4lInit() {
  video_capability cap;
  video_picture pict;
  video_window win;
  char *devname;

  for ( int i = 0; i < 8; i++ ) {
  	asprintf( &devname, "/dev/video%d", i );
	video_fd = open( devname, O_RDWR );
	if ( video_fd != -1 ) {
		printf( "using %s\n", devname );
		free( devname );
		break;
	}

	printf( "open %s %s\n", devname, strerror(errno) );
	free( devname );
  }

  if ( video_fd == -1 ) return false;

  if ( ioctl( video_fd, VIDIOCGCAP, &cap ) == -1 ) {
	perror( "CamView GCAP" );
  	::close( video_fd );
	return false;
  }
//  video_xsize = cap.maxwidth;
//  video_ysize = cap.maxheight;
  video_xsize = 320;
  video_ysize = 240;
//  if ( video_xsize > cap.maxwidth ) video_xsize = cap.maxwidth;
//  if ( video_ysize > cap.maxheight ) video_ysize = cap.maxheight;

  if ( ioctl( video_fd, VIDIOCGWIN, &win ) == -1 ) {
  	perror( "CamView GWIN" );
	::close( video_fd );
	return false;
  }

  int fps = (win.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
  if ( win.width != video_xsize || win.height != video_ysize || fps != 30 ) {
  	win.width = video_xsize;
	win.height = video_ysize;
	win.flags = 30 << PWC_FPS_SHIFT;
  	if ( ioctl( video_fd, VIDIOCSWIN, &win ) == -1 ) {
		perror( "CamView SWIN" );
		::close( video_fd );
		return false;
	}
  }

  if ( ioctl( video_fd, VIDIOCGPICT, &pict ) == -1 ) {
	perror( "CamView GPICT" );
  	::close( video_fd );
	return false;
  }

  if ( pict.palette != VIDEO_PALETTE_YUV420P &&
			pict.palette != VIDEO_PALETTE_RGB24 ) {
  	pict.palette = VIDEO_PALETTE_YUV420P;
	if ( ioctl( video_fd, VIDIOCSPICT, &pict ) != -1 ) goto palette_set;
  	
	pict.palette = VIDEO_PALETTE_RGB24;
	if ( ioctl( video_fd, VIDIOCSPICT, &pict ) == -1 ) {
  		::close( video_fd );
		puts( "no suitable palette found" );
		return false;
	}
palette_set:;
  }

  if ( pict.palette == VIDEO_PALETTE_YUV420P )
	lookup_yuv420p = true;
  else if ( pict.palette == VIDEO_PALETTE_RGB24 ) {
	lookup_rgb24 = true;
	rgb_mode = true;
  }

  vmmap.width = video_xsize;
  vmmap.height = video_ysize;
  vmmap.format = pict.palette;

  if ( ioctl( video_fd, VIDIOCGMBUF, &vmbuf ) == -1 ) {
  	::close( video_fd );
	return false;
  }
  video_map = (char*) mmap( 0, vmbuf.size, PROT_READ | PROT_WRITE,
  	MAP_SHARED, video_fd, 0 );
  if ( video_map == MAP_FAILED ) {
  	perror( "CamView mmap" );
  	::close( video_fd );
  	return false;
  }

  if ( rgb_mode )
	bufsize = video_xsize * video_ysize * 4;
  else
	bufsize = video_xsize * video_ysize * 3 / 2;

  buf0 = (char*) malloc( bufsize );
  
  int gain;
  ioctl( video_fd, VIDIOCPWCSAGC, &gain );
  if ( gain < 0 ) gain = -gain;
  ctrl_panel->gainSL->setValue( gain );

// white balance should be set
  pwc_whitebalance wb_mode;
  wb_mode.mode = 1;
  ioctl( video_fd, VIDIOCPWCSAWB, &wb_mode );

  return true;
}

void CamView::v4lCleanup() {
  free( buf0 );
  munmap( video_map, vmbuf.size );
  ::close( video_fd );
}

bool CamView::xvInit() {
  XvAdaptorInfo *adp_info;
  int i, j, k;
  unsigned int adp_num;

  disp = QPaintDevice::x11AppDisplay();
  XvQueryAdaptors( disp, DefaultRootWindow(disp), &adp_num, &adp_info );

  XvEncodingInfo *enc_info;
  unsigned int enc_count;

  for ( k = 0; k< adp_num; k++ ) {
	for ( i = 0; i < adp_info[k].num_ports; i++ ) {
	        XvQueryEncodings( disp, adp_info[k].base_id + i, &enc_count, &enc_info );
        	for ( j = 0; j < enc_count; j++ )
	                if ( !strcmp( enc_info[j].name, "XV_IMAGE" ) ) {
	                        img_port = adp_info[k].base_id + i;
				goto img_port_found;
			}
	}
	continue;

img_port_found:
	XvImageFormatValues *fmt;
	int fmt_count;
	fmt = XvListImageFormats( disp, img_port, &fmt_count );
	for ( i = 0; i < fmt_count; i++ ) {
		if ( lookup_yuv420p && !strcmp( fmt[i].guid, "I420" ) ) {
			fmt_id = fmt[i].id;
			goto fmt_found;
		}
		else if ( lookup_rgb24 && fmt[i].id == 3 ) {
			fmt_id = 3;
			goto fmt_found;
		}
	}
  }
  puts( "no suitable device for selected palette found" );
  return false;

fmt_found:

  XSetWindowAttributes attr;
  memset( &attr, 0, sizeof attr );
  attr.background_pixel = 0;
  attr.backing_store = Always;
  wout = XCreateWindow( disp,  winId(), 0, 0, width(), height(), 0, 24,
	 CopyFromParent, CopyFromParent, CWBackPixel | CWBackingStore, &attr );
  XMapWindow( disp, wout );
  wout_width = width();
  wout_height = height();

  gc = XCreateGC( disp, wout, 0, 0 );
  printf( "XvImage port=0x%x, fmt=0x%x\n", img_port, fmt_id );
//  fmt_id = 0x30323449;
  imgout = XvCreateImage( disp, img_port, fmt_id, 0, video_xsize, video_ysize );
  if ( !imgout ) {
          puts( "CamView.cc: imgout=0" );
          return false;
  }

  memset( buf0, 0, bufsize );

  return true;
}

void CamView::xvCleanup() {
  XFree( imgout );
  XFreeGC( disp, gc );
  XDestroyWindow( disp, wout );
}

void CamView::paintEvent( QPaintEvent *e ) {
  if ( !initialized ) {
  	QWidget::paintEvent( e );
  	return;
  }
  imgout->data = buf0;
  XvPutImage( disp, img_port, wout, gc, imgout, 0, 0, video_xsize, video_ysize,
  	0, 0, wout_width, wout_height );
  XSync( disp, False );
}

void CamView::resizeEvent( QResizeEvent *e ) {
  if ( !initialized ) {
  	QWidget::resizeEvent( e );
	return;
  }
  if ( panel_shown ) {
	XResizeWindow( disp, wout, width(), height() - ctrl_panel->height() );
  	ctrl_panel->move( 0, height() - ctrl_panel->height() );
	wout_height = height() - ctrl_panel->height();
  }
  else {
  	XResizeWindow( disp, wout, width(), height() );
	ctrl_panel->move( 0, height() );
	wout_height = height();
  }
  wout_width = width();
  ctrl_panel->move( 0, height() - ctrl_panel->height() );
  ctrl_panel->resize( width(), ctrl_panel->height() );
}

void CamView::mousePressEvent( QMouseEvent *e ) {
  bool needs_fixing;

  if ( !initialized || e->button() != RightButton ) return;

  panel_shown = !panel_shown;

  if ( minimumSize() != QSize(0,0) ) {
  	needs_fixing = true;
  	setMinimumSize( 0, 0 );
	setMaximumSize( 32767, 32767 );
	needs_fixing = true;
  }
  else
  	needs_fixing = false;

  if ( panel_shown )
  	resize( width(), height() + ctrl_panel->height() );
  else
	resize( width(), height() - ctrl_panel->height() );

  if ( needs_fixing ) setFixedSize( size() );
}

void CamView::captureLoop() {
  int val;
  unsigned char *s;
  unsigned int *d;

  vmmap.frame = 0;
  while ( !isHidden() ) {
  	if ( ioctl( video_fd, VIDIOCMCAPTURE, &vmmap ) == -1 )
		perror( "ioctl CAPTURE" );
	repaint( false );
	qApp->processEvents();
	val = 0;
	if ( ioctl( video_fd, VIDIOCSYNC, &val ) == -1 )
		perror( "ioctl SYNC" );
	if ( rgb_mode ) {
		s = (unsigned char*) (video_map + vmbuf.offsets[0]);
		d = (unsigned int*) buf0;
		for ( int i = 0; i < video_xsize * video_ysize; i++ ) {
			*d++ = s[0] | (s[1] << 8) | (s[2] << 16);
			s += 3;
		}
	}
	else
		memcpy( buf0, video_map + vmbuf.offsets[0], bufsize );
  }
}

void CamView::disableScale( bool disable ) {
  if ( disable ) {
  	if ( panel_shown )
		resize( video_xsize, video_ysize + ctrl_panel->height() );
	else
		resize( video_xsize, video_ysize );
	setFixedSize( size() );
  }
  else {
  	setMinimumSize( 0, 0 );
	setMaximumSize( 32767, 32767 );
  }
}

void CamView::keepRatio( bool keep ) {
}

void CamView::setRate( int num ) {
  int fps_line[7] = { 5, 10, 15, 20, 25, 30, 0 };
  int i;
  video_window win;

  if ( ioctl( video_fd, VIDIOCGWIN, &win ) == -1 ) {
  	perror( "CamView GWIN" );
	::close( video_fd );
	return;
  }
  fps_line[6] = fps_line[num];

  for ( i = 6; i >= 0; i-- ) {
	win.flags = fps_line[i] << PWC_FPS_SHIFT;
	if ( ioctl( video_fd, VIDIOCSWIN, &win ) == -1 ) {
		perror( "CamView SWIN" );
		::close( video_fd );
		return;
	}

	ioctl( video_fd, VIDIOCGWIN, &win );
	if ( (win.flags >> PWC_FPS_SHIFT) == fps_line[i] ) {
		if ( i != 6 ) ctrl_panel->rateCB->setCurrentItem( i );
		return;
	}
  }

  printf( "unable to change fps, now camera is configured to %d fps\n",
  	win.flags >> PWC_FPS_SHIFT );
}

int CamView::tryResolution( int new_xsize, int new_ysize ) {
  video_window win;
  int fps_line[7] = { 5, 10, 15, 20, 25, 30, 0 };
  int i;

  if ( ioctl( video_fd, VIDIOCGWIN, &win ) == -1 ) {
  	perror( "CamView GWIN" );
	::close( video_fd );
	return -1;
  }
  if ( win.width == new_xsize && win.height == new_ysize ) return 1;

  fps_line[6] = win.flags >> PWC_FPS_SHIFT;

  for ( i = 6; i >= 0; i-- ) {
	win.width = new_xsize;
	win.height = new_ysize;
	win.flags = fps_line[i] << PWC_FPS_SHIFT;
	if ( ioctl( video_fd, VIDIOCSWIN, &win ) == -1 ) {
		perror( "CamView SWIN" );
		::close( video_fd );
		return -1;
	}

	ioctl( video_fd, VIDIOCGWIN, &win );
	if ( win.width == new_xsize && win.height == new_ysize ) {
		if ( i != 6 ) ctrl_panel->rateCB->setCurrentItem( i );
		return 0;
	}
  }
  
  printf( "unsupported mode %dx%d\n", new_xsize, new_ysize );
  win.width = video_xsize;
  win.height = video_ysize;
  win.flags = fps_line[6];
  ioctl( video_fd, VIDIOCSWIN, &win );

  return -1;
}

void CamView::setResolution( int num ) {
  int new_xsize, new_ysize;

  switch ( num ) {
  case 0:
  	new_xsize = 160;
	new_ysize = 120;
	break;
  case 1:
  	new_xsize = 320;
	new_ysize = 240;
	break;
  case 2:
  	new_xsize = 640;
	new_ysize = 480;
	break;
  }

  if ( tryResolution( new_xsize, new_ysize ) != 0 ) return;

  video_xsize = new_xsize;
  video_ysize = new_ysize;
  
  vmmap.width = video_xsize;
  vmmap.height = video_ysize;

  munmap( video_map, vmbuf.size );

  if ( ioctl( video_fd, VIDIOCGMBUF, &vmbuf ) == -1 ) {
  	::close( video_fd );
	return;
  }

  video_map = (char*) mmap( 0, vmbuf.size, PROT_READ | PROT_WRITE,
  	MAP_SHARED, video_fd, 0 );
  if ( video_map == MAP_FAILED ) {
  	perror( "CamView mmap" );
  	::close( video_fd );
  	return;
  }

  XFree( imgout );
  imgout = XvCreateImage( disp, img_port, fmt_id, 0, video_xsize, video_ysize );
  if ( !imgout ) {
          puts( "CamView.cc: imgout=0" );
          return;
  }

  if ( rgb_mode )
	bufsize = video_xsize * video_ysize * 4;
  else
	bufsize = video_xsize * video_ysize * 3 / 2;
  buf0 = (char*) realloc( buf0, bufsize );
}

void CamView::setGain( int gain ) {
  if ( ctrl_panel->autoGainCB->isChecked() ) return;
  ioctl( video_fd, VIDIOCPWCSAGC, &gain );
}

void CamView::setAutoGain( bool auto_gain ) {
  int gain;

  if ( auto_gain ) gain = -65535; else gain = ctrl_panel->gainSL->value();

  ioctl( video_fd, VIDIOCPWCSAGC, &gain );
}

int main( int argc, char **argv ) {
  QApplication qa( argc, argv );

  CamView w;

  if ( w.initialized ) {
	qa.setMainWidget( &w );
	w.captureLoop();
  }
}
