#ifndef _PELOADER_H_
#define _PELOADER_H_

struct pe_header {
  char signature[4];
  short cputype, num_objects;
  int timestamp, resrvd1;
  int resrvd2;
  short nthdr_size, flags;
  short resrvd3, lmajor, lminor, resrvd4;
  int resrvd5, resrvd6;
  int entrypoint_rva, resrvd7;
  int resrvd8, image_base;
  int object_align, file_align;
  short os_major, os_minor, user_major, user_minor;
  short subsys_major, subsys_minor;
  int resrvd9;
  int image_size, header_size;
  int file_checksum;
  short subsystem, dll_flags;
  int stack_reserve_size, stack_commit_size;
  int heap_reserve_size, heap_commit_size;
  int resrvd10, num_rva_sizes;
  int export_table_rva, export_data_size;
  int import_table_rva, import_data_size;
  int resource_table_rva, resource_data_size;
  int exception_table_rva, exception_data_size;
  int security_table_rva, security_data_size;
  int fixup_table_rva, fixup_data_size;
  int debug_table_rva, debug_directories;
  int image_description_rva, image_description_size;
  int machine_specific_rva, machine_specific_size;
  int tls_rva, tls_size;
};

struct pe_obj_entry {
  char name[8];
  int virtual_size;
  int rva;
  int physical_size;
  int physical_offset;
  int resrvd1, resrvd2, resrvd3;
  int object_flags;
};

enum SRB_COMMAND {
  SRB_READ_DATA,
  SRB_WRITE_DATA,
  SRB_GET_STREAM_STATE,			/* 32a */
  SRB_SET_STREAM_STATE,			/* 3ba */
  SRB_SET_STREAM_PROPERTY,
  SRB_GET_STREAM_PROPERTY,		/* 3b2 */
  SRB_OPEN_MASTER_CLOCK,
  SRB_INDICATE_MASTER_CLOCK,		/* 3aa */
  SRB_UNKNOWN_STREAM_COMMAND,
  SRB_SET_STREAM_RATE,
  SRB_PROPOSE_DATA_FORMAT,		/* 3a0 */
  SRB_CLOSE_MASTER_CLOCK,
  SRB_PROPOSE_STREAM_RATE,
  SRB_SET_DATA_FORMAT,			/* 38d */
  SRB_GET_DATA_FORMAT,			/* 31e */
  SRB_BEGIN_FLUSH,
  SRB_END_FLUSH,

  SRB_GET_STREAM_INFO = 0x100,
  SRB_OPEN_STREAM,
  SRB_CLOSE_STREAM,
  SRB_OPEN_DEVICE_INSTANCE,
  SRB_CLOSE_DEVICE_INSTANCE,
  SRB_GET_DEVICE_PROPERTY,
  SRB_SET_DEVICE_PROPERTY,
  SRB_INITIALIZE_DEVICE,
  SRB_CHANGE_POWER_STATE,
  SRB_UNINITIALIZE_DEVICE,
  SRB_UNKNOWN_DEVICE_COMMAND,
  SRB_PAGING_OUT_DRIVER,
  SRB_GET_DATA_INTERSECTION,
  SRB_INITIALIZATION_COMPLETE,
  SRB_SURPRISE_REMOVAL
};

struct HW_CLOCK_OBJECT {
  int HwClockFunction;
  unsigned int ClockSupportFlags;
  unsigned int Reserved[2];
};

struct HW_STREAM_OBJECT {
  unsigned int SizeOfThisPacket;		/* 00 */
  unsigned int StreamNumber;			/* 04 */
  void *HwStreamExtension;			/* 08 */
  int ReceiveDataPacket;			/* 0c */
  int ReceiveControlPacket;			/* 10 */
  HW_CLOCK_OBJECT HwClockObject;		/* 14 */
  char Dma;					/* 24 */
  char Pio;					/* 25 */
  void *HwDeviceExtension;			/* 26 */
  unsigned int StreamHeaderMediaSpecific;	/* 2a */
  unsigned int StreamHeaderWorkspace;		/* 2e */
  char Allocator;				/* 32 */
  int HwEventRoutine;				/* 33 */
  unsigned int Reserved[2];			/* 37 */
  						/* 3f */
};

enum KSSTATE {
  KSSTATE_STOP,
  KSSTATE_ACQUIRE,
  KSSTATE_PAUSE,
  KSSTATE_RUN
};

enum DEVICE_POWER_STATE {
  PowerDeviceUnspecified = 0,
  PowerDeviceD0,
  PowerDeviceD1,
  PowerDeviceD2,
  PowerDeviceD3,
  PowerDeviceMaximum
};

struct KSTIME {
  long long Time;
  unsigned int Numerator;
  unsigned int Denominator;
};

struct KSSTREAM_HEADER {
  unsigned int Size;				/* 00 */
  unsigned int TypeSpecificFlags;		/* 04 */
  KSTIME PresentationTime;			/* 08 */
  long long Duration;				/* 18 */
  unsigned int FrameExtent;			/* 20 */
  unsigned int DataUsed;			/* 24 */
  void *Data;					/* 28 */
  unsigned int OptionsFlags;			/* 2c */
};

struct KSDATAFORMAT {
  unsigned int FormatSize;			/* 00 */
  unsigned int Flags;				/* 04 */
  unsigned int SampleSize;			/* 08 */
  unsigned int Reserved;			/* 0c */
  char MajorFormat[16];		// GUID		/* 10 */
  char SubFormat[16];		// GUID		/* 20 */
  char Specifier[16];		// GUID		/* 30 */
//  char junk1[0x34];				/* 40 */
//  int  magic1, magic2;			/* 74 */
//  char junk2[8];				/* 7c */
//  int magic3;					/* 84 */
};

struct RECT {
  int left;
  int top;
  int right;
  int bottom;
};

struct KS_BITMAPINFOHEADER {
  int biSize;				/* 00 */
  int biWidth;				/* 04 */
  int biHeight;				/* 08 */
  int biPlanes;				/* 0c */
  int biBitCount;			/* 10 */
  int biCompression;			/* 14 */
  int biSizeImage;			/* 18 */
  int biXPelsPerMeter;			/* 1c */
  int biYPelsPerMeter;			/* 20 */
  int biClrUsed;			/* 24 */
  int biClrImportant;			/* 28 */
};

struct KS_VIDEOINFOHEADER {
  RECT rcSource;			/* 00 */
  RECT rcTarget;			/* 10 */
  int dwBitRate;			/* 20 */
  int dwBitErrorRate;			/* 24 */
  long long AvgTimePerFrame;		/* 28 */
  KS_BITMAPINFOHEADER bmiHeader;	/* 30 */
  					/* 58 */
};

struct KS_DATAFORMAT_VIDEOINFOHEADER {
  struct KSDATAFORMAT DataFormat;		/* 00 */
  struct KS_VIDEOINFOHEADER VideoInfoHeader;	/* 40 */
};

struct DEVICE_OBJECT {
#if 1
  char filler[0x50];
#else
  short Type;					/* 00 */
  unsigned short Size;				/* 02 */
  int ReferenceCount;				/* 04 */
  void *DriverObject;				/* 08 */
  void *NextDevice;				/* 0c */
  void *AttachedDevice;				/* 10 */
  void *CurrentIrp;				/* 14 */
  Timer;					/* 18 */
  unsigned int Flags;				/* 1c */
  unsigned int Characteristics;			/* 20 */
  void *DoNotUse1;				/* 24 */
  void *DeviceExtension;			/* 28 */
  DeviceType;					/* 2c */
  char StackSize;				/* 30 */
  union {
	ListEntry;
	Wcb;
  } Queue;
  unsigned int AlignmentRequirement;
  DeviceQueue;
  Dpc;
  unsigned int ActiveThreadsCount;
  void *SecurityDescriptor;
  DeviceLock;
  unsigned short SectorSize;
  unsigned short Spare1;
  void *DeviceObjectExtension;
  void *Reserved;
#endif
};

struct PORT_CONFIGURATION_INFORMATION {
  unsigned int SizeOfThisPacket;		/* 00 */
  void *HwDeviceExtension;			/* 04 */
  DEVICE_OBJECT *ClassDeviceObject;		/* 08 */
  DEVICE_OBJECT *PhysicalDeviceObject;		/* 0c */
  unsigned int SystemIoBusNumber;		/* 10 */
  int AdapterInterfaceType;			/* 14 */
  unsigned int BusInterruptLevel;		/* 18 */
  unsigned int BusInterruptVector;		/* 1c */
  int InterruptMode;				/* 20 */
  unsigned int DmaChannel;			/* 24 */
  unsigned int NumberOfAccessRanges;		/* 28 */
  void *AccessRanges;
  unsigned int StreamDescriptorSize;
  void *Irp;
  void *InterruptObject;
  void *DmaAdapterObject;
  unsigned int Reserved[2];
};

struct HW_STREAM_HEADER {
  unsigned int NumberOfStreams;
  unsigned int SizeOfHwStreamInformation;
  unsigned int NumDevPropArrayEntries;
  void *DevicePropertiesArray;
};

enum KSPIN_DATAFLOW {
  KSPIN_DATAFLOW_IN = 1,
  KSPIN_DATAFLOW_OUT,
  KSPIN_DATAFLOW_FULLDUPLEX
};

struct HW_STREAM_INFORMATION {
  unsigned int NumberOfPossibleInstances;	/* 00 */
  KSPIN_DATAFLOW DataFlow;			/* 04 */
  bool DataAccessible;				/* 08 */
  unsigned int NumberOfFormatArrayEntries;	/* 0c */
  KSDATAFORMAT *StreamFormatsArray;		/* 10 */
  void *ClassReserved[4];			/* 14 */
  unsigned int NumStreamPropArrayEntries;	/* 24 */
  void *StreamPropertiesArray;			/* 28 */
  unsigned int NumStreamEventArrayEntries;	/* 2c */
  void *StreamEventsArray;			/* 30 */
  char *Category;	// GUID*		/* 34 */
  char *Name;		// GUID*		/* 38 */
  unsigned int MediumsCount;			/* 3c */
  void *Mediums;				/* 40 */
  bool BridgeStream;				/* 44 */
  unsigned int Reserved[2];			/* 48 */
  						/* 50 */
};

struct HW_STREAM_DESCRIPTOR {
  HW_STREAM_HEADER StreamHeader;
  HW_STREAM_INFORMATION StreamInformation;
};

struct HW_STREAM_REQUEST_BLOCK {
  unsigned int SizeOfThisPacket;		/* 00 */
  SRB_COMMAND Command;				/* 04 */
  int Status;					/* 08 */
  HW_STREAM_OBJECT *StreamObject;		/* 0c */
  void *HwDeviceExtension;			/* 10 */
  void *SRBExtension;				/* 14 */

  union _CommandData {				/* 18 */
  	KSSTREAM_HEADER *DataBufferArray;
	HW_STREAM_DESCRIPTOR *StreamBuffer;
	KSSTATE StreamState;
	void *TimeReference;
	void *PropertyInfo;
	KSDATAFORMAT *OpenFormat;
	PORT_CONFIGURATION_INFORMATION *ConfigInfo;
	int MasterClockHandle;
	DEVICE_POWER_STATE DeviceState;
	void *IntersectInfo;
  } CommandData;

  unsigned int NumberOfBuffers;			/* 1c */
  unsigned int TimeoutCounter;			/* 20 */
  unsigned int TimeoutOriginal;			/* 24 */
  HW_STREAM_REQUEST_BLOCK *NextSRB;		/* 28 */
  void *Irp;					/* 2c */
  unsigned int flags;
  void *HwInstanceExtension;

  union {
  	unsigned int NumberOfBytesToTransfer;
	unsigned int ActualBytesTransferred;
  };

  void *ScatterGatherBuffer;
  unsigned int NumberOfPhysicalPages;
  unsigned int NumberOfScatterGatherElements;
  unsigned int Reserved[1];
};

struct HW_INITIALIZATION_DATA {
  int HwInitializationDataSize;
  int HwInterrupt;
  int HwReceivePacket;
  int HwCancelPacket;
  int HwRequestTimeoutHandler;
  unsigned int DeviceExtensionSize;
  unsigned int PerRequestExtensionSize;
  unsigned int PerStreamExtensionSize;
  unsigned int FilterInstanceExtensionSize;
  char BusMasterDMA;
  char Dma24BitAddresses;
  unsigned int BufferAlignment;
  char TurnOffSynchronization;
  unsigned int DmaBufferSize;
  unsigned int Reserved[2];
};

enum STREAM_MINIDRIVER_DEVICE_NOTIFICATION_TYPE {
  ReadyForNextDeviceRequest,
  DeviceRequestComplete,
  SignalMultipleDeviceEvents,
  SignalDeviceEvent,
  DeleteDeviceEvent,
  DeviceNotificationMaximum
};

enum POOL_TYPE {
  NonPagedPool,
  PagedPool,
  NonPagedPoolMustSucceed,
  DontUseThisType,
  NonPagedPoolCacheAligned,
  PagedPoolCacheAligned,
  NonPagedPoolCacheAlignedMustS,
  MaxPoolType
};

struct UNICODE_STRING {
  unsigned short Length;
  unsigned short MaximumLength;
  unsigned short *Buffer;
};

struct RTL_QUERY_REGISTRY_TABLE {
  void *QueryRoutine;
  unsigned int Flags;
  unsigned short *Name;
  void *EntryContext;
  unsigned int DefaultType;
  void *DefaultData;
  unsigned int DefaultLength;
};

struct LIST_ENTRY {
  LIST_ENTRY *Flink;
  LIST_ENTRY *Blink;
};

struct URB;

struct USB_CONFIGURATION_DESCRIPTOR {
  unsigned char bLength;			/* 00 */
  unsigned char bDescriptorType;		/* 01 */
  unsigned short wTotalLength;			/* 02 */
  unsigned char bNumInterfaces;			/* 04 */
  unsigned char bConfigurationValue;		/* 05 */
  unsigned char iConfiguration;			/* 06 */
  unsigned char bmAttributes;			/* 07 */
  unsigned char MaxPower;			/* 08 */
  						/* 09 */
};

#define USB_SUBMIT_URB 0x220003
#define USB_CONFIGURATION_DESCRIPTOR_TYPE         0x02
#define URB_BULK_TRANSFER 9

struct URB_HEADER {
  unsigned short Length;			/* 00 */
  unsigned short Function;			/* 02 */
  int Status;					/* 04 */
  void *UsbdDeviceHandle;			/* 08 */
  unsigned int UsbdFlags;			/* 0c */
  						/* 10 */
};

enum USBD_PIPE_TYPE {
  UsbdPipeTypeControl,
  UsbdPipeTypeIsochronous,
  UsbdPipeTypeBulk,
  UsbdPipeTypeInterrupt
};

struct USBD_PIPE_INFORMATION {
  unsigned short MaximumPacketSize;		/* 00 */
  unsigned char EndpointAddress;		/* 02 */
  unsigned char Interval;			/* 03 */
  USBD_PIPE_TYPE PipeType;			/* 04 */
  void *PipeHandle;				/* 08 */
  unsigned int MaximumTransferSize;		/* 0c */
  unsigned int PipeFlags;			/* 10 */
  						/* 14 */
};

struct USBD_INTERFACE_INFORMATION {
  unsigned short Length;			/* 00 */
  unsigned char InterfaceNumber;		/* 02 */
  unsigned char AlternateSetting;		/* 03 */
  unsigned char Class;				/* 04 */
  unsigned char SubClass;			/* 05 */
  unsigned char Protocol;			/* 06 */
  unsigned char Reserved;			/* 07 */
  void *InterfaceHandle;			/* 08 */
  unsigned int NumberOfPipes;			/* 0c */
  USBD_PIPE_INFORMATION Pipes[0];		/* 10 */
  						/* 10 */
};

struct URB_SELECT_CONFIGURATION {
  URB_HEADER Hdr;						/* 00 */
  USB_CONFIGURATION_DESCRIPTOR *ConfigurationDescriptor;	/* 10 */
  void *ConfigurationHandle;					/* 14 */
  USBD_INTERFACE_INFORMATION InterfaceInformation;		/* 18 */
  								/* 28 */
};

struct URB_HCD_AREA {
  void *HcdEndpoint;
  int HcdIrp;
  LIST_ENTRY HcdListEntry;
  LIST_ENTRY HcdListEntry2;
  void *HcdCurrentIoFlushPointer;
  void *HcdExtension;
};

struct URB_BULK_OR_INTERRUPT_TRANSFER {
  URB_HEADER Hdr;
  void *PipeHandle;
  unsigned int TransferFlags;
  unsigned int TransferBufferLength;
  void *TransferBuffer;
  int TransferBufferMDL;
  URB *UrbLink;
  URB_HCD_AREA hcd;
};

struct URB {
  union {
  	URB_HEADER UrbHeader;
  	URB_SELECT_CONFIGURATION UrbSelectConfiguration;
	URB_BULK_OR_INTERRUPT_TRANSFER UrbBulkOrInterruptTransfer;
  };
};

struct IO_STACK_LOCATION {
  unsigned char MajorFunction;
  unsigned char MinorFunction;
  unsigned char Flags;
  unsigned char Control;

  union {
  	struct {
		unsigned int OutputBufferLength;
		unsigned int InputBufferLength;
		unsigned int IoControlCode;
		void *Type3InputBuffer;
	} DeviceIoControl;
	struct {
		void *Argument1;
		void *Argument2;
		void *Argument3;
		void *Argument4;
	} Others;
  } Parameters;

  DEVICE_OBJECT *DeviceObject;
  int FileObject;
  int CompletionRoutine;
  int Context;
};

struct IO_STATUS_BLOCK {
  int Status;
  unsigned int Information;
};

enum _int_WObjType {
  ObjTimer,
  ObjEvent,
  ObjSemaphore
};

struct DISPATCHER_HEADER {
  unsigned char Type;
  unsigned char Absolute;
  unsigned char Size;
  unsigned char Inserted;
  int SignalState;
  LIST_ENTRY WaitListHead;
};

struct KDPC {
  short Type;
  unsigned char Number;
  unsigned char Importance;
  LIST_ENTRY DpcListEntry;
  void *DeferredRoutine;
  void *DeferredContext;
  void *SystemArgument1;
  void *SystemArgument2;
  unsigned int *Lock;
};

struct KTIMER {
  DISPATCHER_HEADER Header;
  long long DueTime;
  LIST_ENTRY TimerListEntry;
  KDPC *Dpc;
  int Period;
};

struct KEVENT {
  DISPATCHER_HEADER Header;
};

struct KSEMAPHORE {
  DISPATCHER_HEADER Header;
  int Limit;
};

struct WORK_QUEUE_ITEM {
  LIST_ENTRY List;
  int WorkerRoutine;
  void *Parameter;
};

#endif /* _PELOADER_H_ */
